/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
#include "hls_aes.h"
#include "hls_utils.h"
using namespace hls;

#define TAU 16

enum fsms_t {S_INIT,
    S_AD, S_MSG,
    S_TAG, S_OUT};

/* -------------------------------------------------------*/

#if !__SYNTHESIS__
    #define DBG 1
#endif  // if !__SYNTHESIS__
#define SYNTHESIS_FAIL 0

void gf128_mul2(array_t b, array_t a)
{
#pragma HLS INLINE
  data_t atmp;
  atmp = reshape_array(a);
  atmp = (atmp << 1);
  reshape_block(b, atmp);
  if ((a[0] >> 7) & 0x1)
    b[15] ^= 0x87;
}

void gf128_mul3(array_t c, array_t a)
{
#pragma HLS INLINE
  array_t   tmp;
#pragma HLS ARRAY_RESHAPE variable=a            complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c            complete dim=1
#pragma HLS ARRAY_RESHAPE variable=tmp            complete dim=1
  gf128_mul2(tmp, a);
  xor2(c, a, tmp);
}


void gf128_mul7(array_t b, array_t a)
{
#pragma HLS INLINE
  array_t ax2, ax6;
#pragma HLS ARRAY_RESHAPE variable=ax2          complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ax6          complete dim=1
#pragma HLS ARRAY_RESHAPE variable=a            complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b            complete dim=1
  gf128_mul2(ax2, a);
  gf128_mul3(ax6, ax2);
  xor2(b, ax6, a);
}


/* -------------------------------------------------------*/

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE ap_vld  port=msg_auth_valid
#if __SYNTHESIS__
#else
    size_t dbg_loop = 0;
#endif

    size_t i;
    public_bus bdi;
    output_bus bdo;
    secret_bus key;


    bool loop = 1;
    bool is_odd = 1;
    bool is_first = 1;
    fsms_t fsm = S_INIT;
    static array_t key_reg;
    array_t sum, L, Ls, output_r, input_r;
    array_t add;
    array_t rk;
    array_t new_rk;
    array_t add_enc;
    array_t TA;
#pragma HLS ARRAY_RESHAPE variable=key_reg complete dim=1
#pragma HLS ARRAY_RESHAPE variable=sum complete dim=1
#pragma HLS ARRAY_RESHAPE variable=L complete dim=1
#pragma HLS ARRAY_RESHAPE variable=Ls complete dim=1
#pragma HLS ARRAY_RESHAPE variable=output_r complete dim=1
#pragma HLS ARRAY_RESHAPE variable=input_r  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_rk complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add_enc complete dim=1
#pragma HLS ARRAY_RESHAPE variable=TA complete dim=1


    // Key initialization (if needed)
    if (key_update == 1) {
        LOAD_KEY: for (i=0; i<4; i++) {
            key_bus.read(key);
            key_reg[i*4+0] = (key.data >> 24) & 0xFF;
            key_reg[i*4+1] = (key.data >> 16) & 0xFF;
            key_reg[i*4+2] = (key.data >>  8) & 0xFF;
            key_reg[i*4+3] = (key.data >>  0) & 0xFF;
        }
    }

    // Initialize here
    clear(sum);
    clear(L);
    clear(Ls);

    bdi_bus.read(bdi);
    if (bdi.type == TYPE_AD) {
        fsm = S_INITQ;
    } else { // TYPE_NPUB
        fsm = S_NPUB;
    }    

    while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 20) return;
        else dbg_loop++;
#endif
        array_t to_aes, from_aes;
        array_t to_aes_xor, input;
        array_t sel_from_aes, sel_input;
        data_bytes_t sel_vld, sel_pad;
        
#if DBG_AES
        std::cout << "Current fsm = " << fsm << std::endl;
        std::cout << "\tInput: " << std::hex << std::uppercase << std::setw(16) << reshape_array(to_aes) << " " << reshape_array(state+8) << std::endl;
        std::cout << "\tKey  : " << std::hex << std::uppercase << std::setw(16) << reshape_array(key_reg) << " " << reshape_array(key_reg+8) << std::endl;
#endif
        
        // Top
        if (S_AD_SUM == fsm || S_DATA_SUM == fsm) {
            copy(input, sum);
        } else {
            reshape_block(input, bdi.data);
        }
        
        if (S_NPUB == fsm)
            input[15] ^= TAU;
        
        if ((S_AD == fsm || S_DATA == fsm) && bdi.eot && ~(bdi.valid_bytes & 0x1)) {
            sel_vld = bdi.valid_bytes;
            sel_pad = bdi.pad_loc;
        } else {
            sel_vld = 0xFFFF;
            sel_pad = 0;
        }        
        
        // L
        array_t sel_L1, sel_L2;
        array_t sel_L1x2, sel_L1x3, sel_L1x7, sel_L1x9
        if (S_AD == fsm || (S_DATA == fsm && is_odd))
            copy(sel_L1, L);
        else
            copy(sel_L1, Ls); 
        gf128_mul2(sel_L1x2, sel_L1);
        gf128_mul3(sel_L1x3, sel_L1);
        gf128_mul3(sel_L1x9, sel_L1);
        gf128_mul7(sel_L1x7, sel_L1);
        if (S_AD == fsm && bdi.eot && (bdi.valid_bytes & 0x1)) {
            copy(sel_L2, sel_L1x9);
        } else if (S_DATA_SUM == fsm && !is_odd && (bdi.valid_bytes & 0x1)) {
            copy(sel_L2, sel_1x7);
        } else if ((S_AD == fsm && bdi.eot) || (S_DATA_SUM == fsm && !is_odd)) {
            copy(sel_L2, sel_L1x3);        
        } else {
            copy(sel_L2, sel_L1);
        }
                           
              
        // Main input
        byte_valid_pad(sel_input, input, sel_vld, sel_pad);                      
        
        if (S_INITQ == fsm || S_NPUB == fsm)
            copy(to_aes_xor, sel_input);
        else
            xor2(to_aes_xor, sel_from_aes, sel_input);
        
                
        xor2(to_aes, to_aes_xor, sel_L2);
        
        AddRoundKey(add, to_aes, key_reg);
        AESKeyUpdate(rk, key_reg, 0);
        
        // Output
        array_t to_output1, to_output2;
        if (is_odd) 
            copy(to_output1, input_r);
        else 
            copy(to_output1, to_aes_xor);
        
        if (S_DATA_SUM == fsm) 
            copy(to_output2, input_r);
        else 
            copy(to_output2, to_aes_xor);
        
        if ((S_DATA == fsm && !is_first) || (S_DATA_SUM == fsm)) {
            bdo.data = reshape_array(to_output2);
            bdo_bus.write(bdo);
        }
        
        // Update register
        if (S_AD == fsm || (S_DATA == fsm && is_odd))
            copy(L, sel_L1x2);
        if (S_DATA == fsm && !is_odd)
            copy(Ls, sel_L1x2);               
        if (S_AD == fsm && !is_first) 
            xor2(sum, from_aes, sum);
        else if (S_DATA == fsm)
            xor2(sum, sel_input, sum);

        if (S_DATA == fsm && is_odd)
            reshape_block(input_r, bdi.data);
        if (S_DATA == fsm && is_odd)
            copy(output_r, to_output1);
        /* ========= */
        /* MAIN LOOP */
        /* ========= */
        AES_LOOP:
        for (i=0; i<AES_ROUND; i++) {
#if DBG_ROUND
            std::cout << "\tAES Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(add) << " " << reshape_array(add+8) << std::endl;
            std::cout << "\tKey Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(rk)  << " " << reshape_array(rk+8) << std::endl;
#endif
            AESKeyUpdate(new_rk, rk, i+1);
            AESRound(from_aes, add_enc, add, rk);
            copy(add, add_enc);
            copy(rk, new_rk);
        }

#if DBG_AES
        std::cout << "\tOutpt: " << std::hex << std::uppercase << std::setw(16) << reshape_array(state) << reshape_array(state+8) << std::endl;
#endif

        /* ========= */
        /* Post-processing  */
        /* ========= */
        switch (fsm) {
        case S_INITQ:
            copy(L, from_aes);
            fsm = S_AD;
        case S_AD:
            if (bdi.eot) fsm = S_AD_SUM;
            else bdi_bus.read(bdi);
            break;
        case S_AD_SUM:
            bdi_bus.read(bdi);
            copy(TA, from_aes);
            break;
        case S_NPUB:
            copy(L, from_aes);
            gf128_mul3(Ls, from_aes);                        
            bdi_bus.read(bdi);
            fsm = S_DATA;
            break;            
        case S_DATA:
            if (bdi.eot) fsm = S_DATA_SUM;
            else bdi_bus.read(bdi);
            is_odd = !is_odd;
            is_first = 0;
            break;
        case S_DATA_SUM:
            loop = 0;            
        }
    }
    
    // Tag
    public_bus bdi_tmp;
    data_t exp_tag; 
    array_t tag;   
#pragma HLS ARRAY_RESHAPE variable=tag complete dim=1    
    
    xor2(tag, TA, from_aes);
    
    if (decrypt) {                
        bdi_bus.read(bdi_tmp);
        reshape_block(exp_tag, bdi_tmp);
        *msg_auth_valid = (exp_tag == tag) ? 1:0;
    } else {
        bdo.data = reshape_array(tag);
        bdo_bus.write(bdo);
    }
    
    return;
}
